import logging
from dataclasses import dataclass, field
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import List

import aiohttp

from tastyworks.models.security import Security

LOGGER = logging.getLogger(__name__)


class OrderType(Enum):
    LIMIT = 'Limit'
    MARKET = 'Market'


class OrderPriceEffect(Enum):
    CREDIT = 'Credit'
    DEBIT = 'Debit'


class OrderStatus(Enum):
    RECEIVED = 'Received'
    CANCELLED = 'Cancelled'
    FILLED = 'Filled'
    EXPIRED = 'Expired'
    LIVE = 'Live'
    REJECTED = 'Rejected'

    def is_active(self):
        return self in (OrderStatus.LIVE, OrderStatus.RECEIVED)


class TimeInForce(Enum):
    DAY = 'Day'
    GTC = 'GTC'
    GTD = 'GTD'


@dataclass
class OrderDetails(object):
    type: OrderType = None
    time_in_force: TimeInForce = TimeInForce.DAY
    gtc_date: datetime = None
    price: Decimal = None
    price_effect: OrderPriceEffect = None
    status: OrderStatus = None
    legs: List[Security] = field(default_factory=list)
    source: str = 'WBT'

    def is_executable(self) -> bool:
        required_data = all([
            self.time_in_force,
            self.price_effect,
            self.price is not None,
            self.type,
            self.source
        ])

        if not required_data:
            return False

        if not self.legs:
            return False

        if self.time_in_force == TimeInForce.GTD:
            try:
                datetime.strptime(self.gtc_date, '%Y-%m-%d')
            except ValueError:
                return False

        return True


class Order(Security):
    def __init__(self, order_details: OrderDetails):
        """
        Initiates a new order object.

        Args:
            order_details (OrderDetails): An object specifying order-level details.
        """
        self.details = order_details

    def check_is_order_executable(self):
        return self.details.is_executable()

    def add_leg(self, security: Security):
        self.details.legs.append(security)

    @classmethod
    def from_dict(cls, input_dict: dict):
        """
        Parses an Order object from a dict.
        """
        details = OrderDetails(input_dict['underlying-symbol'])
        details.price = Decimal(input_dict['price']) if 'price' in input_dict else None
        details.price_effect = OrderPriceEffect(input_dict['price-effect'])
        details.type = OrderType(input_dict['order-type'])
        details.status = OrderStatus(input_dict['status'])
        details.time_in_force = input_dict['time-in-force']
        details.gtc_date = input_dict.get('gtc-date', None)
        return cls(order_details=details)

    @classmethod
    async def get_remote_orders(cls, session, account, **kwargs) -> List:
        """
        Gets all orders on Tastyworks.

        Args:
            session (TastyAPISession): The session to use.
            account (TradingAccount): The account_id to get orders on.
            Keyword arguments specifying filtering conditions, these include:
            `status`, `time-in-force`, etc.

        Returns:
            list(Order): A list of Orders
        """
        if not session.logged_in:
            raise Exception('Tastyworks session not logged in.')

        filters = kwargs
        url = '{}/accounts/{}/orders'.format(
            session.API_url,
            account.account_number
        )
        url = '{}?{}'.format(
            url,
            '&'.join([f'{k}={v}' for k, v in filters.items()])
        )

        res = []
        async with aiohttp.request('GET', url, headers=session.get_request_headers()) as resp:
            if resp.status != 200:
                raise Exception('Could not get current open orders')
            data = (await resp.json())['data']['items']
            for order_data in data:
                order = cls.from_dict(order_data)
                if not order.details.status.is_active():
                    continue
                res.append(order)
        return res



import datetime
import logging

import requests

LOGGER = logging.getLogger(__name__)


class TastyAPISession(object):
    def __init__(self, username: str, password: str, API_url=None):
        self.API_url = API_url if API_url else 'https://api.tastyworks.com'
        self.username = username
        self.password = password
        self.logged_in = False
        self.session_token = self._get_session_token()

    def _get_session_token(self):
        if self.logged_in and self.session_token:
            if (datetime.datetime.now() - self.logged_in_at).total_seconds() < 60:
                return self.session_token

        body = {
            'login': self.username,
            'password': self.password
        }
        resp = requests.post(f'{self.API_url}/sessions', json=body)
        if resp.status_code != 201:
            self.logged_in = False
            self.logged_in_at = None
            self.session_token = None
            raise Exception('Failed to log in, message: {}'.format(resp.json()['error']['message']))

        self.logged_in = True
        self.logged_in_at = datetime.datetime.now()
        self.session_token = resp.json()['data']['session-token']
        self._validate_session()
        return self.session_token

    def is_active(self):
        return self._validate_session()

    def _validate_session(self):
        resp = requests.post(f'{self.API_url}/sessions/validate', headers=self.get_request_headers())
        if resp.status_code != 201:
            self.logged_in = False
            self.logged_in_at = None
            self.session_token = None
            raise Exception('Could not validate the session, error message: {}'.format(
                resp.json()['error']['message']
            ))
            return False
        return True

    def get_request_headers(self):
        return {
            'Authorization': self.session_token
        }



import aiohttp

from tastyworks.dough import BASE_URL


async def stock_search(symbols):
    """
    Performs a stock search through the dough API.

    Args:
        symbols (list): The list of string symbols to search for.

    Returns:
        list (dict): A list of stock details including full name, implied volatility, liquidity etc.

    Raises:
        Exception: raises a general exception when the request does not succeed.
    """

    request_url = '{}/stocks/search/'.format(
        BASE_URL
    )

    payload = {
        'symbols': symbols
    }

    async with aiohttp.request('POST', request_url, json=payload) as resp:
        if resp.status != 200:
            raise Exception(
                'Searching for stocks {} failed'.format(', '.join(payload['symbols']))
            )
        data = await resp.json()

    return data["stocks"]



from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal


@dataclass
class Greeks(object):
    symbol: str = None
    index: str = None
    datetime: datetime = None
    price: Decimal = None
    volatility: Decimal = None
    delta: Decimal = None
    gamma: Decimal = None
    theta: Decimal = None
    rho: Decimal = None
    vega: Decimal = None

    def from_streamer_dict(self, input_dict: dict):
        """
        imports the Greeks data from a dictionary pulled from subscribed streamer data
            sub_greeks = {"Greeks": [".SPY210419P410"]}
            await streamer.add_data_sub(sub_greeks)

        Args:
            input_dict (dict): dictionary from the streamer containing the greeks data for one options symbol

        """
        self.symbol = input_dict.get('eventSymbol')
        self.index = input_dict.get('index')
        self.datetime = datetime.fromtimestamp(input_dict.get('time') / 1000.0)  # timestamp comes in ms
        self.price = Decimal(str(input_dict.get('price')))
        self.volatility = Decimal(str(input_dict.get('volatility')))
        self.delta = Decimal(str(input_dict.get('delta')))
        self.gamma = Decimal(str(input_dict.get('gamma')))
        self.theta = Decimal(str(input_dict.get('theta')))
        self.rho = Decimal(str(input_dict.get('rho')))
        self.vega = Decimal(str(input_dict.get('vega')))

        return self



import datetime

from tastyworks.dxfeed.mapped_item import MappedItem


class Trade(MappedItem):
    DXFEED_TEXT = 'Trade'

    def _process_fields(self, data_dict: dict):
        data_dict['time'] = datetime.datetime.fromtimestamp(data_dict['time'] / 1000_000_000)
        return data_dict

    def __init__(self, data=None):
        super().__init__(data=data)



from dataclasses import dataclass
from datetime import date
from decimal import Decimal
from enum import Enum

from tastyworks.models.greeks import Greeks
from tastyworks.models.security import Security
from tastyworks.models.underlying import UnderlyingType


class OptionType(Enum):
    PUT = 'P'
    CALL = 'C'


@dataclass
class Option(Security):
    ticker: str
    expiry: date
    strike: Decimal
    option_type: OptionType
    underlying_type: UnderlyingType
    greeks: Greeks = None
    quantity: int = 1
    symbol_occ: str = None
    symbol_dxf: str = None

    def __post_init__(self):
        self.symbol_occ = self.get_occ2010_symbol()
        self.symbol_dxf = self.get_dxfeed_symbol()

    def _get_underlying_type_string(self, underlying_type: UnderlyingType):
        if underlying_type == UnderlyingType.EQUITY:
            return 'Equity Option'

    def get_occ2010_symbol(self):
        strike_int, strike_dec = divmod(self.strike, 1)
        strike_int = int(round(strike_int, 5))
        strike_dec = int(round(strike_dec, 3) * 1000)

        res = '{ticker}{exp_date}{type}{strike_int}{strike_dec}'.format(
            ticker=self.ticker[0:6].ljust(6),
            exp_date=self.expiry.strftime('%y%m%d'),
            type=self.option_type.value,
            strike_int=str(strike_int).zfill(5),
            strike_dec=str(strike_dec).zfill(3)
        )
        return res

    def get_dxfeed_symbol(self):
        if self.strike % 1 == 0:
            strike_str = '{0:.0f}'.format(self.strike)
        else:
            strike_str = '{0:.2f}'.format(self.strike)
            if strike_str[-1] == '0':
                strike_str = strike_str[:-1]

        res = '.{ticker}{exp_date}{type}{strike}'.format(
            ticker=self.ticker,
            exp_date=self.expiry.strftime('%y%m%d'),
            type=self.option_type.value,
            strike=strike_str
        )
        return res

    def to_tasty_json(self):
        res = {
            'instrument-type': f'{self.underlying_type.value} Option',
            'symbol': self.get_occ2010_symbol(),
            'quantity': self.quantity
        }
        return res



import aiohttp

from tastyworks.models.session import TastyAPISession


async def symbol_search(symbol: str, session: TastyAPISession):
    """
    Performs a symbol search using Tastyworks API.

    This returns a list of symbols that are similar to the symbol passed in
    the parameters. This does not provide any details except the related
    symbols and their descriptions.

    Args:
        symbol (string): A base symbol to search for similar symbols

    Returns:
        list (dict): A list of symbols and descriptions that are closely
                     related to the passed symbol parameter
    """

    url = f'{session.API_url}/symbols/search/{symbol}'

    async with aiohttp.request('GET', url, headers=session.get_request_headers()) as resp:
        if resp.status != 200:
            raise Exception(
                f'Failed to query symbols. Response status: {resp.status}; message: {resp.json()["error"]["message"]}'
            )
        data = await resp.json()

    return data['data']['items']



from enum import Enum

from tastyworks.models.security import Security


class UnderlyingType(Enum):
    EQUITY = 'Equity'


class Underlying(Security):
    def __init__(self, ticker=None):
        self.ticker = ticker



import calendar
from datetime import date, timedelta

VERSION = '4.2.1'


def get_third_friday(d):
    s = date(d.year, d.month, 15)
    candidate = s + timedelta(days=(calendar.FRIDAY - s.weekday()) % 7)

    # This month's third friday passed
    if candidate < d:
        candidate += timedelta(weeks=4)
        if candidate.day < 15:
            candidate += timedelta(weeks=1)

    return candidate



import logging
from datetime import date, datetime
from decimal import Decimal
from typing import Dict

import aiohttp

from tastyworks.models.option import Option, OptionType
from tastyworks.models.underlying import Underlying, UnderlyingType

LOGGER = logging.getLogger(__name__)


class OptionChain(object):
    """
    Maps option symbols to Option structures
    Provides filter methods based on type, strike price, expiry, etc.

    Example usage:
        symbol = "AAPL"
        response = await client.get_option_chains(symbol)
        oc = OptionChain(OptionChain.parse(symbol, response))

        for strike in oc.get_all_strikes():
            print(strike)

        for expiry in oc.get_all_expirations():
            print(expiry)

        for symbol, option in oc.get(strike=300.0,option_type=OptionType.CALL,expiry=datetime.date(2018,9,21)).items():
            print(symbol, option)
    """

    def __init__(self, options):
        self.options = options

    def _get_filter_strategy(self, key, unique=True):
        values = [getattr(option, key) for option in self.options]
        if not any(values):
            raise Exception(f'No values found for specified key: {key}')

        values = list(set(values)) if unique else list(values)
        return sorted(values)

    def get_all_strikes(self):
        return self._get_filter_strategy('strike')

    def get_all_expirations(self):
        return self._get_filter_strategy('expiry')


async def get_option_chain(session, underlying: Underlying, expiration: date = None) -> OptionChain:
    LOGGER.debug('Getting options chain for ticker: %s', underlying.ticker)
    data = await _get_tasty_option_chain_data(session, underlying)
    res = []

    for exp in data['expirations']:
        exp_date = datetime.strptime(exp['expiration-date'], '%Y-%m-%d').date()

        if expiration and expiration != exp_date:
            continue

        for strike in exp['strikes']:
            strike_val = Decimal(strike['strike-price'])
            for option_types in OptionType:
                new_option = Option(
                    ticker=underlying.ticker,
                    expiry=exp_date,
                    strike=strike_val,
                    option_type=option_types,
                    underlying_type=UnderlyingType.EQUITY
                )
                res.append(new_option)
    return OptionChain(res)


async def _get_tasty_option_chain_data(session, underlying) -> Dict:
    async with aiohttp.request(
            'GET',
            f'{session.API_url}/option-chains/{underlying.ticker}/nested',
            headers=session.get_request_headers()) as response:

        if response.status != 200:
            raise Exception(f'Could not find option chain for symbol {underlying.ticker}')
        resp = await response.json()

        # NOTE: Have not seen an example with more than 1 item. No idea what that would be.
        return resp['data']['items'][0]



import logging
import sys

log = logging.getLogger(__name__)
log.propagate = False
out_hdlr = logging.StreamHandler(sys.stdout)
out_hdlr.setFormatter(logging.Formatter('%(asctime)s %(message)s'))
out_hdlr.setLevel(logging.INFO)
log.addHandler(out_hdlr)
log.setLevel(logging.INFO)

root = logging.getLogger()
root.addHandler(out_hdlr)
root.propagate = False
root.setLevel(logging.INFO)



import aiohttp

BASE_URL = 'https://trade.tastyworks.com'


async def search_symbol(query):
    request_url = '{}/symbol_search/search/{}'.format(BASE_URL, query)
    async with aiohttp.request('GET', request_url) as resp:
        if resp.status != 200:
            raise Exception('Could not search for symbol {}', query)
        resp = await resp.json()
        return [entry[0] for entry in resp]



import aiohttp

from tastyworks.models.session import TastyAPISession


class Watchlist(object):
    """Watchlist Class"""

    def __init__(self):
        """init"""
        self.name = None
        self.group_name = None
        self.securities = {}

    @classmethod
    def from_list(cls, data: list):
        """From List

        Class Factory

        Notes:
            There is an instability issue with Tastyworks API.  Hence
            the need to catch an exceptions.  Special symbols don't have
            instrument type's at all.

            For stability & accessibility, all instrument type's are returned
            as 'instrument_type'.

        Args:
            data (list):   List of dict objects containing symbol &
                           instrument_type

        Returns:
            Watchlist:  Instance of Watchlist()
        """
        inst = cls()
        for item in data:
            try:
                inst.securities[item['symbol']] = {
                    'instrument_type': item['instrument-type']
                }
            except KeyError:
                try:
                    inst.securities[item['symbol']] = {
                        'instrument_type': item['instrument_type']
                    }
                except KeyError:
                    inst.securities[item['symbol']] = {
                        'instrument_type': None
                    }
        return inst

    def __str__(self):
        """String Representation"""
        return str(self.__dict__)


class WatchlistGroup(object):
    """WatchlistGroup Class"""

    def __init__(self):
        """init"""
        self.watchlists = {}

    def __iter__(self):
        """Iterator"""
        return iter(self.watchlists)

    def __getitem__(self, item):
        """Subscriptor"""
        return self.watchlists[item]

    def __repr__(self):
        """Object Representation"""
        return str(self.watchlists)

    def __str__(self):
        """String Representation"""
        return str(list(self.watchlists.keys()))

    @classmethod
    async def get_watchlists(
        cls, session: TastyAPISession, public: bool = True
    ):
        """Get Watchlists

        Class Factory

        Retrieves a watchlist group (either public or private).

        Notes:
            Not all Watchlist's have group names.

        Args:
            session (TastyAPISession): A TastyAPISession object
            public (bool): Retrive public or private watchlists

        Returns:
            WatchlistGroup: Instance of WatchlistGroup()
        """
        url = f'{session.API_url}/public-watchlists' if public else f'{session.API_url}/watchlists'

        async with aiohttp.request('GET', url, headers=session.get_request_headers()) as resp:
            if resp.status != 200:
                raise Exception(
                    f'Failed retrieving watchlists, Response status: {resp.status}; message: {resp.json()["error"]["message"]}'
                )
            data = await resp.json()

        data = data['data']['items']
        inst = cls()
        for entry in data:
            list_data = entry['watchlist-entries']
            wlist = Watchlist.from_list(list_data)
            wlist.name = entry['name']
            try:
                wlist.group_name = entry['group-name']
            except KeyError:
                pass
            inst.watchlists[wlist.name] = wlist

        return inst



import datetime
import logging

import aiocometd
import requests
from aiocometd import ConnectionType

from tastyworks import dxfeed
from tastyworks.dxfeed import mapper as dxfeed_mapper
from tastyworks.models.session import TastyAPISession

LOGGER = logging.getLogger(__name__)


class DataStreamer(object):
    def __init__(self, session: TastyAPISession):
        if not session.is_active():
            raise Exception('TastyWorks API session not active/valid')
        self.tasty_session = session
        self.cometd_client = None
        self.subs = {}

    @classmethod
    async def create(cls, session: TastyAPISession):
        self = DataStreamer(session)
        await self._setup_connection()
        return self

    async def add_data_sub(self, values):
        LOGGER.debug(f'Adding subscription: {values}')
        await self._send_msg(dxfeed.SUBSCRIPTION_CHANNEL, {'add': values})

    async def remove_data_sub(self, values):
        # NOTE: Experimental, unconfirmed. Needs testing
        LOGGER.info(f'Removing subscription: {values}')
        await self._send_msg(dxfeed.SUBSCRIPTION_CHANNEL, {'remove': values})

    async def _consumer(self, message):
        return dxfeed_mapper.map_message(message)

    async def _send_msg(self, channel, message):
        if not self.logged_in:
            raise Exception('Connection not made or logged in')
        LOGGER.debug('[dxFeed] sending: %s on channel: %s', message, channel)
        await self.cometd_client.publish(channel, message)

    async def reset_data_subs(self):
        LOGGER.debug('Resetting data subscriptions')
        await self._send_msg(dxfeed.SUBSCRIPTION_CHANNEL, {'reset': True})

    def get_streamer_token(self):
        return self._get_streamer_data()['data']['token']

    def _get_streamer_data(self):
        if not self.tasty_session.logged_in:
            raise Exception('Logged in session required')

        if hasattr(self, 'streamer_data_created') and (datetime.datetime.now() - self.streamer_data_created).total_seconds() < 60:
            return self.streamer_data

        resp = requests.get(f'{self.tasty_session.API_url}/quote-streamer-tokens', headers=self.tasty_session.get_request_headers())
        if resp.status_code != 200:
            raise Exception('Could not get quote streamer data, error message: {}'.format(
                resp.json()['error']['message']
            ))
        self.streamer_data = resp.json()
        self.streamer_data_created = datetime.datetime.now()
        return resp.json()

    def _get_streamer_websocket_url(self):
        socket_url = self._get_streamer_data()['data']['websocket-url']
        full_url = '{}/cometd'.format(socket_url)
        return full_url

    async def _setup_connection(self):
        aiocometd.client.DEFAULT_CONNECTION_TYPE = ConnectionType.WEBSOCKET
        streamer_url = self._get_streamer_websocket_url()
        LOGGER.info('Connecting to url: %s', streamer_url)

        auth_extension = AuthExtension(self.get_streamer_token())
        cometd_client = aiocometd.Client(
            streamer_url,
            auth=auth_extension,
        )
        await cometd_client.open()
        await cometd_client.subscribe(dxfeed.DATA_CHANNEL)

        self.cometd_client = cometd_client
        self.logged_in = True
        LOGGER.info('Connected and logged in to dxFeed data stream')

        await self.reset_data_subs()

    async def listen(self):
        async for msg in self.cometd_client:
            LOGGER.debug('[dxFeed] received: %s', msg)
            if msg['channel'] != dxfeed.DATA_CHANNEL:
                continue
            yield await self._consumer(msg['data'])


class AuthExtension(aiocometd.AuthExtension):
    def __init__(self, streamer_token: str):
        self.streamer_token = streamer_token

    def _get_login_msg(self):
        return {'ext': {'com.devexperts.auth.AuthToken': f'{self.streamer_token}'}}

    def _get_advice_msg(self):
        return {
            'timeout': 60 * 1000,
            'interval': 0
        }

    async def incoming(self, payload, headers=None):
        pass

    async def outgoing(self, payload, headers=None):
        for entry in payload:
            if 'clientId' not in entry:
                entry.update(self._get_login_msg())

    async def authenticate(self):
        pass



import abc
import logging

from tastyworks.dxfeed import mapper as mapper

LOGGER = logging.getLogger(__name__)


class MappedItem(object):
    __metaclass__ = abc.ABCMeta

    DXFEED_TEXT = None

    def _map_data(self, data) -> list:
        first_sample = True
        if isinstance(data[0], str):
            first_sample = False

        if first_sample:
            if data[0][0] != self.DXFEED_TEXT:
                raise Exception('Input JSON data does not contain a quote')
        else:
            if data[0] != self.DXFEED_TEXT:
                raise Exception('Input JSON data does not contain a quote')

        if first_sample:
            keys = data[0][1]
            # NOTE: I know this is dirty. Technical debt.
            # Stores the list of keys from the first sample since
            # subsequent ones only provide values
            mapper.KEY_MAP[self.DXFEED_TEXT] = keys
        else:
            keys = mapper.KEY_MAP[self.DXFEED_TEXT]

        res = []
        values = data[1]
        # if we have a 'multi-sample' i.e. several items subscribed to
        multiples = len(values) / len(keys)
        if not multiples.is_integer():
            raise Exception('Mapper data input values are not an integer multiple of the key size')
        for i in range(int(multiples)):
            offset = i * len(keys)
            local_values = values[offset:(i + 1) * len(keys)]
            res.append(self._process_fields(dict(zip(keys, local_values))))
        return res

    def _process_fields(self, data_dict: dict):
        """
        Used to post-process fields in the element's dictionary.
        e.g. convert Unix time to datetimes
        """
        return data_dict

    def __init__(self, data=None):
        if data:
            self.data = self._map_data(data)
        self.keys = None



import logging

from tastyworks.dxfeed import greeks, profile, quote, summary, trade

LOGGER = logging.getLogger(__name__)
KEY_MAP = {}


def map_message(message):
    if isinstance(message[0], str):
        first_sample = False
    else:
        first_sample = True
    msg_type = message[0][0] if first_sample else message[0]

    if quote.Quote.DXFEED_TEXT == msg_type:
        res = quote.Quote(data=message)
    elif greeks.Greeks.DXFEED_TEXT == msg_type:
        res = greeks.Greeks(data=message)
    elif trade.Trade.DXFEED_TEXT == msg_type:
        res = trade.Trade(data=message)
    elif summary.Summary.DXFEED_TEXT == msg_type:
        res = summary.Summary(data=message)
    elif profile.Profile.DXFEED_TEXT == msg_type:
        res = profile.Profile(data=message)
    else:
        LOGGER.warning("Unknown message type received from streamer: {}".format(message))
        res = [{'warning': 'Unknown message type received', 'message': message}]

    return res



from dataclasses import dataclass
from typing import List

import aiohttp

from tastyworks.models.order import Order, OrderPriceEffect


@dataclass
class TradingAccount(object):
    account_number: str
    external_id: str
    is_margin: bool
    is_closed: bool
    account_type_name: str
    nickname: str
    opened_at: str

    async def execute_order(self, order: Order, session, dry_run=True):
        """
        Execute an order. If doing a dry run, the order isn't placed but simulated (server-side).

        Args:
            order (Order): The order object to execute.
            session (TastyAPISession): The tastyworks session onto which to execute the order.
            dry_run (bool): Whether to do a test (dry) run.

        Returns:
            bool: Whether the order was successful.
        """
        if not order.check_is_order_executable():
            raise Exception('Order is not executable, most likely due to missing data')

        if not session.is_active():
            raise Exception('The supplied session is not active and valid')

        url = '{}/accounts/{}/orders'.format(
            session.API_url,
            self.account_number
        )
        if dry_run:
            url = f'{url}/dry-run'

        body = _get_execute_order_json(order)

        async with aiohttp.request('POST', url, headers=session.get_request_headers(), json=body) as resp:
            if resp.status == 201:
                return True
            elif resp.status == 400:
                raise Exception('Order execution failed, message: {}'.format(await resp.text()))
            else:
                raise Exception('Unknown remote error, status code: {}, message: {}'.format(resp.status, await resp.text()))

    @classmethod
    def from_dict(cls, data: dict) -> List:
        """
        Parses a TradingAccount object from a dict.
        """
        new_data = {
            'is_margin': True if data['margin-or-cash'] == 'Margin' else False,
            'account_number': data['account-number'],
            'external_id': data['external-id'],
            'is_closed': True if data['is-closed'] else False,
            'account_type_name': data['account-type-name'],
            'nickname': data['nickname'],
            'opened_at': data['opened-at']
        }

        res = TradingAccount(**new_data)

        return res

    @classmethod
    async def get_remote_accounts(cls, session) -> List:
        """
        Gets all trading accounts from the Tastyworks platform.

        Args:
            session (TastyAPISession): An active and logged-in session object against which to query.

        Returns:
            list (TradingAccount): A list of trading accounts.
        """
        url = f'{session.API_url}/customers/me/accounts'
        res = []

        async with aiohttp.request('GET', url, headers=session.get_request_headers()) as response:
            if response.status != 200:
                raise Exception('Could not get trading accounts info from Tastyworks...')
            data = (await response.json())['data']

        for entry in data['items']:
            if entry['authority-level'] != 'owner':
                continue
            acct_data = entry['account']
            acct = TradingAccount.from_dict(acct_data)
            res.append(acct)

        return res

    async def get_balance(self, session, **kwargs):
        """
        Get balance.

        Args:
            session (TastyAPISession): An active and logged-in session object against which to query.
        Returns:
            dict: account attributes
        """
        url = '{}/accounts/{}/balances'.format(
            session.API_url,
            self.account_number
        )

        async with aiohttp.request('GET', url, headers=session.get_request_headers(), **kwargs) as response:
            if response.status != 200:
                raise Exception('Could not get trading account balance info from Tastyworks...')
            data = (await response.json())['data']
        return data

    async def get_positions(self, session, **kwargs):
        """
        Get Open Positions.

        Args:
            session (TastyAPISession): An active and logged-in session object against which to query.
        Returns:
            dict: account attributes
        """
        url = '{}/accounts/{}/positions'.format(
            session.API_url,
            self.account_number
        )

        async with aiohttp.request('GET', url, headers=session.get_request_headers(), **kwargs) as response:
            if response.status != 200:
                raise Exception('Could not get open positions info from Tastyworks...')
            data = (await response.json())['data']['items']
        return data

    async def get_live_orders(self, session, **kwargs):
        """
        Get live Orders.

        Args:
            session (TastyAPISession): An active and logged-in session object against which to query.
        Returns:
            dict: account attributes
        """
        url = '{}/accounts/{}/orders/live'.format(
            session.API_url,
            self.account_number
        )

        async with aiohttp.request('GET', url, headers=session.get_request_headers(), **kwargs) as response:
            if response.status != 200:
                raise Exception('Could not get live orders info from Tastyworks...')
            data = (await response.json())['data']['items']
        return data

    async def get_history(self, session, **kwargs):
        """
        Get transaction history.

        Args:
            session (TastyAPISession): An active and logged-in session object against which to query.
        Returns:
            dict: account attributes
        """
        url = '{}/accounts/{}/transactions'.format(
            session.API_url,
            self.account_number
        )

        PAGE_SIZE = 1024

        page_offset = 0
        total_pages = None
        total_items = 0
        all_items = []

        while total_pages is None or page_offset < total_pages:
            # print(f'Getting page {page_offset}')
            params = kwargs.pop('params', {})
            params.update({
                'per-page': PAGE_SIZE,
                'page-offset': page_offset,
            })
            kwargs['params'] = params

            async with aiohttp.request('GET', url, headers=session.get_request_headers(), **kwargs) as response:
                if response.status != 200:
                    raise Exception('Could not get history info from Tastyworks...')
                data = (await response.json())

            page_offset += 1
            if total_pages is None:
                pagination = data['pagination']
                total_pages = pagination['total-pages']
                total_items = pagination['total-items']

            items = data['data']['items']
            if items:
                all_items.extend(items)

        if len(all_items) != total_items:
            raise Exception('Could not fetch some items in paginated request.')

        return all_items


def _get_execute_order_json(order: Order):
    order_json = {
        'source': order.details.source,
        'order-type': order.details.type.value,
        'price': '{:.2f}'.format(order.details.price),
        'price-effect': order.details.price_effect.value,
        'time-in-force': order.details.time_in_force.value,
        'legs': _get_legs_request_data(order)
    }

    if order.details.gtc_date:
        order_json['gtc-date'] = order.details.gtc_date.strftime('%Y-%m-%d')

    return order_json


def _get_legs_request_data(order):
    res = []
    order_effect = order.details.price_effect
    order_effect_str = 'Sell to Open' if order_effect == OrderPriceEffect.CREDIT else 'Buy to Open'
    for leg in order.details.legs:
        leg_dict = {**leg.to_tasty_json(), 'action': order_effect_str}
        res.append(leg_dict)
    return res



from setuptools import find_packages, setup

from tastyworks.utils import VERSION


f = open('README.md', 'r')
LONG_DESCRIPTION = f.read()
f.close()

setup(
    name='tastyworks-api',
    version=VERSION,
    description='An unofficial API for Tastyworks!',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Graeme Holliday',
    author_email='graeme.holliday@pm.me',
    url='https://github.com/Graeme22/tastyworks-api',
    license='Apache',
    install_requires=[
        'aiohttp<4',
        'requests<3',
        'aiocometd',
        'dataclasses'
    ],
    packages=find_packages(exclude=['ez_setup', 'tests*']),
    include_package_data=True,
)
